from typing import Any, Callable, List, Optional, Tuple, TypeVar, Union

from mypy_extensions import NamedArg, VarArg


class InjectorError(Exception):
    """ Base class for dependency injection errors"""


class ProvideError(InjectorError):
    """ Raise this error, when error occured in provide phase"""


class InjectError(InjectorError):
    """ Raise this error, when error occured in injection phase"""


class NoKwOnly(ProvideError):
    """ User raise this error, when no suitable KwOnly argument is available """


FACTORY = 1
SINGLETON = 2
SCOPED = 3
CUSTOM = 4
VALUE = 5


class BoundInjectable:
    def __call__(self) -> Any:
        """ Return provided injectable result

        example::

            injector = Injector()

            class A:
                pass

            injectable = injector.provide(A)
            bound = injectable.bind(injector)
            assert isinstance(bound, BoundInjectable)
            assert isinstance(bound(), A)
        """


class Injectable:
    def bind(self, injector: "Injector") -> BoundInjectable:
        """ Bind injectable to injector """

    # TODO: delete resolve, or KIDERÍTENI MIRE JÓ
    def resolve(self, injector: "Injector") -> Any:
        pass

    def __call__(self, injector: "Injector") -> Any:
        """ Resolve injectable value with the given injector """


Provide_id = object
Provide_value = Optional[Union[type, object, Callable]]
Provide_strategy = Optional[int]
# Provide_provide = Optional[List[Tuple[Provide_id, Provide_value, Provide_strategy, "Provide_provide"]]]
Provide_provide = Optional[List[Tuple[Provide_id, Provide_value, Provide_strategy, Any]]]

CR = TypeVar("CR")


class Injector:
    def provide(self,
                id: Provide_id,
                value: Provide_value = None,
                strategy: Provide_strategy = FACTORY,
                provide: Provide_provide = None) -> Injectable:
        pass

    def get(self, id: Provide_id) -> Any:
        """ Resolve injected value with the given id """

    def exec(self, callable: Callable[..., CR], provide: Provide_provide = None) -> CR:
        """ Executes the callable with the current injector.

        DO NOT USE IN PERFORMANCE CRITICAL SCNARIOS.

        This method is good for unit test, example::

            injector = Injector()

            @injector.provide
            class A:
                pass

            @injector.exec
            def test_something(a: A):
                assert isinstance(a, A)
        """

    def descend(self) -> "Injector":
        """ Returns a new Injector instance and inherits providers from the current Injector """

    def injectable(value: Provide_value = None,
                   strategy: Provide_strategy = FACTORY,
                   provide: Provide_provide = None) -> Injectable:
        """ Returns a new Injectable instance. Useful for caching injectable without provideing """

    def __getitem__(self, id: Provide_id) -> Any:
        """ Resolve injected value with the given id, slightly faster than ``get`` method """


class ValueResolver:
    pass


class KwOnly:
    """ Provide keyword only arguments

    example::
        injector = Injector()

        class Config(dict):
            def __init__(self):
                super().__init__(some_key="OK")

        def get_kwarg(config: Config, *, name, type):
            assert name == "some_key"
            assert type is str
            return config[name]

        def fn(*, some_key: str):
            assert some_key == "OK"
            return "NICE"

        injector.provide(Config)
        injector.provide(KwOnly(get_kwarg))

        assert injector.exec(fn) == "NICE"

    """

    def __init__(self, callable: Callable[[
            VarArg(Any),
            NamedArg(type=str, name="name"),
            NamedArg(type=Any, name="type"),
    ], Any]):
        pass



from typing import Generic, TypeVar, overload, NoReturn

Inj = TypeVar("Inj")
Inst = TypeVar("Inst")


class Inject(Generic[Inj]):
    """Inject token in class level

    Example::
        from yapic.di import Injector, Inject

        class Database:
            pass

        class A:
            db: Inject[Database]

        injector = Injector()
        injector.provide(Database)
        injector.provide(A)

        a_inst = injector.get(A)
        assert(isinstance(a_inst.db, Database))
    """
    @overload
    def __get__(self, instance: None, owner: type) -> NoReturn:
        pass

    @overload
    def __get__(self, instance: Inst, owner: type) -> Inj:  # noqa
        pass

    def __get__(self, instance, owner):  # noqa
        return None



#!/usr/bin/env python3

import os
import sys
from glob import glob
from os import path
from pathlib import Path
from setuptools import setup, Extension
from setuptools.command.test import test as TestCommand
from setuptools import Command

VERSION = "2.0.4"

define_macros = {
    "YAPIC_DI_VERSION_MAJOR": VERSION.split(".")[0],
    "YAPIC_DI_VERSION_MINOR": VERSION.split(".")[1],
    "YAPIC_DI_VERSION_PATCH": VERSION.split(".")[2]
}
undef_macros = []
extra_compile_args = []  # -flto

subcommand_args = []
if "--" in sys.argv:
    subcommand_args = sys.argv[sys.argv.index("--") + 1:]
    del sys.argv[sys.argv.index("--"):]

if sys.platform == "win32":
    define_macros["UNICODE"] = "1"

    DEVELOP = sys.executable.endswith("python_d.exe")
    if DEVELOP:
        define_macros["_DEBUG"] = "1"
        undef_macros.append("NDEBUG")
        extra_compile_args.append("/MTd")
        extra_compile_args.append("/Zi")
    else:
        undef_macros.append("_DEBUG")
        extra_compile_args.append("/Ox")

    extra_compile_args.append("/FAs")
else:
    extra_compile_args.append("-std=c++11")
    extra_compile_args.append("-Wno-unknown-pragmas")
    extra_compile_args.append("-Wno-write-strings")

    DEVELOP = sys.executable.endswith("-dbg")
    if DEVELOP:
        define_macros["_DEBUG"] = 1
        undef_macros.append("NDEBUG")
    else:
        extra_compile_args.append("-O3")


def root(*p):
    return path.join(path.dirname(path.abspath(__file__)), *p)


cpp_ext = Extension(
    name="yapic.di._di",
    sources=["src/di.cpp"],
    include_dirs=["./libs/yapic.core/src/yapic/core/include"],
    depends=glob("src/*.hpp") + glob("./libs/yapic.core/src/yapic/core/include/**/*.hpp"),
    extra_compile_args=extra_compile_args,
    define_macros=list(define_macros.items()),
    undef_macros=undef_macros,
    language="c++",
)


def cmd_prerun(cmd, requirements):
    for r in requirements(cmd.distribution):
        if r:
            installed = cmd.distribution.fetch_build_eggs(r)

            if installed:
                for dp in map(lambda x: x.location, installed):
                    if dp not in sys.path:
                        sys.path.insert(0, dp)

    cmd.distribution.get_command_obj("build").force = True
    cmd.run_command("build")

    ext = cmd.get_finalized_command("build_ext")
    ep = str(Path(ext.build_lib).absolute())

    if ep not in sys.path:
        sys.path.insert(0, ep)

    for e in ext.extensions:
        if e._needs_stub:
            ext.write_stub(ep, e, False)


class PyTest(TestCommand):
    user_options = [
        ("file=", "f", "File to run"),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.file = "./tests/"

    def finalize_options(self):
        super().finalize_options()

    def run(self):
        def requirements(dist):
            yield dist.install_requires
            yield dist.tests_require

        cmd_prerun(self, requirements)
        self.run_tests()

    def run_tests(self):
        import pytest
        errno = pytest.main(subcommand_args)
        sys.exit(errno)


class Benchmark(Command):
    user_options = [
        ("file=", "f", "File to run"),
    ]

    def initialize_options(self):
        self.file = None
        self.pytest_args = "-x -s"

    def finalize_options(self):
        if self.file:
            self.pytest_args += " " + self.file.replace("\\", "/")

    def run(self):
        def requirements(dist):
            yield dist.extras_require["benchmark"]

        cmd_prerun(self, requirements)
        import shlex
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


# typing: https://github.com/python/typing/issues/84
setup(
    name="yapic.di",
    version=VERSION,
    url="https://github.com/zozzz/yapic.di",
    author="Zoltán Vetési",
    author_email="vetesi.zoltan@gmail.com",
    description="Very fast Dependency Injection library",
    long_description=(Path(__file__).parent / "README.rst").read_text(encoding="utf-8"),
    license="BSD",
    packages=["yapic.di"],
    package_dir={"yapic.di": "src"},
    package_data={"yapic.di": ["_di.pyi"]},
    ext_modules=[cpp_ext],
    tests_require=["pytest", "pytest-leaks"],
    python_requires=">=3.7",
    extras_require={"benchmark": ["pytest", "pytest-benchmark"]},
    cmdclass={
        "test": PyTest,
        "bench": Benchmark
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)
